from sqlalchemy import func, and_, select

def argmax(query, maximise_column, group_bys):
    """Returns a modified query that emulates an argmax function for SQL

    :param query: The SQLAlchemy query
    :param maximise_column: The table column to maximise
    :param group_bys: The table columns to group by
    :returns: modified SQLAlchemy query that returns the rows which satisfy argmax(maximise_column)
    """

    # we're going to have a subquery that looks very much like the main query,
    # but which serves to pick out the argmax
    argmax_subquery = query

    argmax_subquery = argmax_subquery.add_columns(func.max(maximise_column).label("maximise_column_result")). \
        group_by(*group_bys).subquery()
    # NB tried making the above a materialised CTE to improve performance, but it doesn't help (presumably the SQLite
    # optimizers already spot the subquery is used twice?)

    join_conditions = [group_by == argmax_subquery.corresponding_column(group_by) for group_by in group_bys]
    # the main query will join to the subquery to pick out all details of the selected row
    query = query.join(argmax_subquery,
                       and_( maximise_column == argmax_subquery.c.maximise_column_result,
                            *join_conditions))

    # there may be rare occasions where two links have exactly the same weight, in which case the above query
    # currently generates more than one row. Avoid by grouping. Note in this case SQL will return
    # basically a random choice of which row to return, so this is not a perfect solution - long term TODO.
    # This also ends up necessitating using MySQL without ONLY_FULL_GROUP_BY mode enabled
    query = query.group_by(*group_bys)

    return query

def delete_non_maximal_rows(connection, table, maximise_column, group_bys):
    """Deletes columns from a table that are *not* the argmax.

    This is provided as a performance workaround because of the high cost of the argmax query

    :param connection: connection on which to execute the delete
    :param table: The SQLAlchemy table
    :param maximise_column: The table column to maximise
    :param group_bys: The table columns to group by
    :returns: number of deleted rows
    """
    max_weight = func.max(maximise_column).label("max_weight")

    source_and_max_weight = select(max_weight, *group_bys).\
        group_by(*group_bys).subquery(name="max_from_source_id")

    group_bys_subquery = [source_and_max_weight.corresponding_column(c) for c in group_bys]
    ids_to_eliminate = select(table.primary_key.columns[0]).\
        join(source_and_max_weight,and_(*[a==b for a,b in zip(group_bys, group_bys_subquery)])).\
        filter(maximise_column < source_and_max_weight.c.max_weight)

    # with sqlite it's OK to just do in_(ids_to_eliminate) but MySQL needs a subquery construction, otherwise
    # it complains it can't do the delete
    ids_to_eliminate_cte = ids_to_eliminate.subquery()
    ids_to_eliminate_select = select(ids_to_eliminate_cte.c.id)

    delete_statement = table.delete(table.primary_key.columns[0].in_(ids_to_eliminate_select))
    deleted_count = connection.execute(delete_statement).rowcount

    return deleted_count


"""Configuration module for tangos

Rather than change anything directly in this file, you can create a config_local.py with the variable you
want to override and it will automatically take precedence.
"""

from __future__ import absolute_import
import os
import sqlalchemy.dialects.mysql

home = os.environ['HOME']
db = os.environ.get("TANGOS_DB_CONNECTION", home+"/tangos_data.db")
base = os.environ.get("TANGOS_SIMULATION_FOLDER", home+"/")

testing_db_user = os.environ.get("TANGOS_TESTING_DB_USER", None)
testing_db_password = os.environ.get("TANGOS_TESTING_DB_PASSWORD", None)
testing_db_backend = os.environ.get("TANGOS_TESTING_DB_BACKEND", 'sqlite')

DOUBLE_PRECISION = sqlalchemy.Float().\
    with_variant(sqlalchemy.dialects.mysql.DOUBLE(asdecimal=False), "mysql")
LARGE_BINARY = sqlalchemy.LargeBinary().\
    with_variant(sqlalchemy.dialects.mysql.LONGBLOB(), "mysql")

default_fileset_handler_class = "pynbody.PynbodyInputHandler"

default_linking_threshold = 0.005
# the percentage of particles in common between two objects before the database bothers to store the relationship

min_halo_particles = 1000
# the minimum number of particles needed in an object before the database bothers to store it

max_num_objects = None
# the maximum number of objects of any one class to store in the database (or None for no limit).
# Normally, the most massive objects will be stored, but this does depend on either the halo finder
# ordering appropriately, or the re-ordering of objects being enabled (which is the default)

default_backend = 'null'
# the default paralellism backend. Set e.g. to mpi4py to avoid having to pass --backend mpi4py to all parallel runs.


# names of property modules to import; default is for backwards compatibility on systems with N-Body-Shop extensions
# N.B. since version 1.0.10 it is also possible to use entry point tangos.property_modules. For example, in your
# my_fab_tangos_properties package setup.py's call to setup() add the keyword argument:
#
#       entry_points={"tangos.property_modules" : [
#           "fab = my_fab_tangos_properties"
#       ]}
#

property_modules = os.environ.get("TANGOS_PROPERTY_MODULES", "tangos_nbodyshop_properties")
property_modules = property_modules.split(",")
property_modules = map(str.strip, property_modules)

file_ignore_pattern = []

max_traverse_depth = 3

# merger tree thinning criteria (applied at query time, not at time of writing links)
mergertree_min_fractional_weight = 0.02 # as a fraction of the weight of the strongest link from each halo
mergertree_min_fractional_NDM = 0.01 # as a fraction of the most massive halo at each timestep - set to zero for no thinning
mergertree_max_nhalos = 30 # maximum number of halos per step - discard the least massive ones
mergertree_timeout = 15.0 # seconds before abandoning the construction of a merger tree in the web interface
mergertree_max_hops = 500 # maximum number of timesteps to scan

# relation finding paremeters for multi hop queries
num_multihops_max_default = 100     # the maximum number of links to follow when searching for related halos
max_relative_time_difference = 1e-4     # the maximum fractional difference in time between two contemporaneous timesteps when searching for related halos

# On some network file systems, concurrency using sqlite is dodgy to say the least. After committing a transaction
# on one node, and before attempting to open a new transaction on another node, it seems empirically helpful to
# allow a significant time delay. This variable controls that delay.
DEFAULT_SLEEP_BEFORE_ALLOWING_NEXT_LOCK = 1.0
# number of seconds to sleep after a lock is released before reallocating it

# Default format to use in the webview. Can be either svg or png
webview_default_image_format = 'svg'

# Caching time in seconds for image and data in the web server
webview_cache_time = 3600

try:
    from .config_local import *
except:
    pass



from __future__ import absolute_import
from pyramid.view import view_config
from sqlalchemy import func, and_, or_
from . import timestep_from_request

import tangos
from tangos import core


@view_config(route_name='timestep_view', renderer='../templates/timestep_view.jinja2')
def timestep_view(request):
    ts = timestep_from_request(request)
    sim = ts.simulation

    all_objects = []

    typecode = 0
    while True:
        try:
            typetag = core.SimulationObjectBase.object_typetag_from_code(typecode)
        except ValueError:
            break

        n_objects = request.dbsession.query(core.SimulationObjectBase).\
            filter_by(timestep_id=ts.id, object_typecode=typecode).order_by(core.SimulationObjectBase.halo_number).count()

        title = core.SimulationObjectBase.class_from_tag(typetag).__name__+"s"

        if title=="BHs":
            title="Black holes"
        elif title=="PhantomHalos":
            title="Phantom halos"

        if n_objects>0:
            all_objects.append({'title': title, 'typetag': typetag, 'n_items': n_objects,
                                'object_url': request.route_url('halo_view',simid=sim.escaped_basename, timestepid=ts.escaped_extension,
                                                                halonumber=typetag+"_")
                                })

        typecode+=1


    return {'timestep': ts.extension,
            'objects': all_objects,
            'timestep_url': request.route_url('timestep_view',simid=request.matchdict['simid'],
                                            timestepid=request.matchdict['timestepid'])}



from .. import TimeChunkedProperty
from . import pynbody_handler_module, PynbodyPropertyCalculation

import numpy as np

class StarFormHistogram(TimeChunkedProperty):
    works_with_handler = pynbody_handler_module.PynbodyInputHandler
    requires_particle_data = True

    @classmethod
    def name(self):
        return "SFR_histogram"

    def plot_xlabel(self):
        return "t/Gyr"

    def plot_ylabel(self):
        return r"$SFR/M_{\odot} yr^{-1}$"

    def calculate(self, halo, existing_properties):
        try:
            weights = halo.st['massform']
        except KeyError:
            weights = halo.st['mass']

        tmax_Gyr = 20.0 # calculate up to 20 Gyr
        nbins = int(20.0/self.pixel_delta_t_Gyr)

        M,_ = np.histogram(halo.st['tform'].in_units("Gyr"),weights=weights.in_units("Msol"),bins=nbins,range=(0,tmax_Gyr))
        t_now = halo.properties['time'].in_units("Gyr")
        M/=self.pixel_delta_t_Gyr
        M = M[self.store_slice(t_now)]

        return M

    def reassemble(self, *options):
        reassembled = super(StarFormHistogram, self).reassemble(*options)
        return reassembled/1e9 # Msol per Gyr -> Msol per yr

class StarForm(PynbodyPropertyCalculation):
    names = "SFR_10Myr", "SFR_100Myr"
    
    def calculate(self, halo, existing_properties):
        halo = halo.star

        t_now = halo.properties['time'].in_units("Gyr")
        tform = halo['tform'].in_units("Gyr")
        mask_10Myr = (t_now-tform)<0.01
        mask_100Myr = (t_now-tform)<0.1
        # Because physical_units has been called previously, mass is in Msol. Return results in Msol/yr.
        return halo['mass'][mask_10Myr].sum()/1e7, halo['mass'][mask_100Myr].sum()/1e8



from __future__ import absolute_import
from __future__ import print_function

from .. import parallel_tasks
from ..log import logger
from .. import core
from . import GenericTangosTool
from ..util import proxy_object
from ..util import timestep_object_cache
import numpy as np
import numbers

class PropertyImporter(GenericTangosTool):
    tool_name = 'import-properties'
    tool_description = 'Import properties that were calculated by the halo finder'

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')

        parser.add_argument('--type', action='store', type=str, dest='typetag', default='halo',
                            help="Specify the object type to run on by tag name (e.g. 'halo' or 'group')")

        parser.add_argument('properties', action='store', nargs='*',
                            help="The names of the halo-finder pre-calculated properties to import; if not specified, all available properties are imported.")

        parser.add_argument('--backwards', action='store_true',
                            help='Process low-z timesteps first')

    def process_options(self, options):
        self.options = options

    def _create_property(self, name, object, value):
        """Create a single database property corresponding to the given value

        See _create_properties for more information."""
        if isinstance(value, proxy_object.ProxyObjectBase):
            value = value.relative_to_timestep_cache(self._object_cache).resolve(self._session)
            if value is not None:
                return core.halo_data.HaloLink(object, value, name)
        elif isinstance(value, numbers.Number):
            return core.halo_data.HaloProperty(object, name, value)
        elif isinstance(value, np.ndarray):
            if np.issubdtype(value.dtype, np.number):
                return core.halo_data.HaloProperty(object, name, value)
            else:
                logger.warn("Ignoring stat file entry key='%s' value='%s' as the value is not a number or an array of numbers",
                        name.text, value)
        elif value is not None:
            logger.warn("Ignoring stat file entry key='%s' value='%s' as the value is not a number or an array of numbers",
                        name.text, value)
        return None

    def _create_properties(self, name, object, values):
        """Create database property or properties corresponding to the given values.

        The values can be proxy objects, to indicate a link should be created

        :arg name: the name ORM object
        :arg object: the object with which the property should be associated
        :arg values: the value, or a list of values
        :returns: a list of objects to be added to the database (always a list, even if there is only one value)
        """

        if isinstance(values, list):
            objects = [self._create_property(name, object, v) for v in values]
        else:
            objects = [self._create_property(name, object, values)]

        return filter(lambda x: x is not None, objects)

    def _import_properties_for_timestep(self, ts, property_names, object_typetag):
        """Import the named properties for a specific timestep

        :arg ts: the database timestep
        :arg property_names: list of names to import, or empty list to import all available names
        :arg object_typetag: the type tag of the objects for which properties will be imported

        :type ts: core.timestep.TimeStep
        """

        logger.info("Processing %s", ts)

        if len(property_names)==0:
            property_names = self.handler.available_object_property_names_for_timestep(ts.extension, object_typetag)

        self._object_cache = timestep_object_cache.TimestepObjectCache(ts)
        self._session = core.Session.object_session(ts)

        property_db_names = [core.dictionary.get_or_create_dictionary_item(self._session, name) for name in
                             property_names]
        rows_to_store = []
        for values in self.handler.iterate_object_properties_for_timestep(ts.extension, object_typetag, property_names):
            db_object = self._object_cache.resolve_from_finder_offset(values[0], object_typetag)
            if db_object is not None:
                for db_name, value in zip(property_db_names, values[2:]):
                    rows_to_store+=self._create_properties(db_name, db_object, value)


        logger.info("Add %d properties", len(rows_to_store))
        self._session.add_all(rows_to_store)
        self._session.commit()

    def run_calculation_loop(self):
        base_sim = core.sim_query_from_name_list(self.options.sims)

        names = self.options.properties
        object_typetag = self.options.typetag
    
        for x in base_sim:
            timesteps = core.get_default_session().query(core.timestep.TimeStep).filter_by(
                simulation_id=x.id, available=True).order_by(core.timestep.TimeStep.redshift.desc()).all()

            if self.options.backwards:
                timesteps = timesteps[::-1]

            self.handler = x.get_output_handler()

            for ts in parallel_tasks.distributed(timesteps):
                self._import_properties_for_timestep(ts, names, object_typetag)



from .. import core, query
from . import GenericTangosTool


class PropertyDeleter(GenericTangosTool):
    tool_name = 'delete-properties'
    tool_description = 'Delete named properties from a halo, timestep, simulation or entire database'
    parallel = False

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--for', '--sims', action='store', nargs='*',
                            metavar='name',
                            help='Specify one or more simulations, timesteps, or halos to run on',
                            dest="for_")

        parser.add_argument('--force', '-f', action='store_true',
                            help='Do not prompt before deleting')

        parser.add_argument('properties', action='store', nargs='+',
                            help="The names of the properties to delete")


    def process_options(self, options):
        self.options = options

    def run_calculation_loop(self):
        dictids = [core.get_dict_id(p) for p in self.options.properties]
        session = core.get_default_session()

        if self.options.for_ is not None:
            base_query = session.query(core.HaloProperty.id).filter(core.HaloProperty.name_id.in_(dictids))
            print(f"Delete {', '.join(self.options.properties)}")
            queries = []
            for s in self.options.for_:
                q = base_query
                obj = query.get_item(s)
                if isinstance(obj, core.Simulation):
                    q = q.join(core.SimulationObjectBase).join(core.TimeStep).join(core.Simulation).filter(core.Simulation.id == obj.id)
                elif isinstance(obj, core.TimeStep):
                    q = q.join(core.SimulationObjectBase).join(core.TimeStep).filter(core.TimeStep.id == obj.id)
                elif isinstance(obj, core.SimulationObjectBase):
                    q = q.join(core.SimulationObjectBase).filter(core.SimulationObjectBase.id == obj.id)

                print(f"  from {obj} ({q.count():d} total properties)")

                # it's not permitted to delete from an expression with a join, so now we need to create a wrapping query
                q = session.query(core.HaloProperty).filter(core.HaloProperty.id.in_(q.from_self()))
                queries.append(q)
        else:
            queries = [session.query(core.HaloProperty).filter(core.HaloProperty.name_id.in_(dictids))]
            print(f"Delete {', '.join(self.options.properties)} from entire database "
                  f"({queries[0].count()} total properties)")

        ok = self.options.force
        if not ok:
            print("""Type "yes" to continue""")
            ok = input(":").lower() == "yes"
        if ok:
            for q in queries:
                q.delete(synchronize_session=False)
            session.commit()
            print("Completed")
        else:
            print("Aborted")




from __future__ import absolute_import
import os
import copy
import numpy as np

from . import translations
from ...util import proxy_object
from six.moves import range
from six.moves import zip


class HaloStatFile(object):
    """Manages and reads a halo stat file of unspecified format."""
    _finder_offset_start = 0 #whether finder_offset should start at 0 (default) or N (typically either 0 or 1)
    _column_translations = {}

    def __new__(cls, timestep):
        subcls = cls.find_loadable_subclass(timestep)
        if subcls:
            return object.__new__(subcls)
        else:
            raise IOError("No stat file found for timestep %r"%timestep)

    @classmethod
    def find_loadable_subclass(cls, timestep_filename):
        if cls.can_load(timestep_filename):
            return cls
        for subclass in cls.__subclasses__():
            loadable_cls = subclass.find_loadable_subclass(timestep_filename)
            if loadable_cls:
                return loadable_cls
        return None

    @classmethod
    def can_load(cls, timestep_filename):
        try:
            return os.path.exists(cls.filename(timestep_filename))
        except (ValueError, TypeError):
            return False

    @classmethod
    def filename(cls, timestep_filename):
        raise ValueError("Unknown path to stat file")

    def __init__(self, timestep_filename):
        self._timestep_filename = timestep_filename
        self.filename = self.filename(timestep_filename)

    def all_columns(self):
        with open(self.filename) as f:
            columns = self._read_column_names(f)

        columns+=self._column_translations.keys()
        return columns

    def iter_rows_raw(self, *args):
        """
        Yield
        1) the index in which each halo appears in the catalog (starting from 0 unless _finder_offset_start is set)
        2) the raw halo ID (finder_id) included in the halo stat file without any emulation
        3) the rest of the requested parameters

        :param args: strings for the column names
        :return: finder_offset_start, finder_id, arg1, arg2, arg3, ... where finder_offset_start is the index of the halo within
        the stat file, finder_id is the raw halo ID read from the stat file, and argN is the value associated with the
        Nth column name provided as input.
        """
        with open(self.filename) as f:
            header = self._read_column_names(f)
            cnt = 0
            ids = [0]
            for a in args:
                try:
                    ids.append(header.index(a))
                except ValueError:
                    ids.append(None)
            for l in f:
                if not l.startswith("#"):
                    col_data = self._get_values_for_columns(ids, l)
                    col_data.insert(0, cnt+self._finder_offset_start)
                    yield col_data
                    cnt += 1

    def iter_rows(self, *args):
        """
        Yield the requested column values from the halo catalog stat file, as well as the finder_offset (index associated
        with halo's position within the catalog) and finder_id (raw halo id listed in the stat file).

        Returned halo properties are emulated when necessary. For example, AHF stat files do not contain n_dm; however,
        its value can be automatically inferred by this function. Meanwhile IDL .amiga.stat files rename n_gas as N_gas.

        :param args: strings for the column names
        :return: finder_offset, finder_id, arg1, arg2, arg3 where argN is the value of the Nth named column
        """

        raw_args = []
        for arg in args:
            if arg in self._column_translations:
                raw_args+=self._column_translations[arg].inputs()
            else:
                raw_args.append(arg)
        for raw_values in self.iter_rows_raw(*raw_args):
            values = [raw_values[0], raw_values[1]]
            for arg in args:
                if arg in self._column_translations:
                    values.append(self._column_translations[arg](raw_args, raw_values[2:]))
                else:
                    values.append(raw_values[2:][raw_args.index(arg)])
            yield values

    def read(self, *args):
        """Read the halo ID and requested columns from the entire file, returning each column as a separate array"""
        return_values = [[] for _ in range(len(args)+2)]
        for row in self.iter_rows(*args):
            for return_array, value in zip(return_values, row):
                return_array.append(value)

        return [np.array(x) for x in return_values]

    def _get_values_for_columns(self, columns, line):
        results = []
        l_split = line.split()
        for id_this in columns:
            if id_this is None:
                this_cast = None
            else:
                this_str = l_split[id_this]
                if "." in this_str or "e" in this_str:
                    guess_type = float
                else:
                    guess_type = int

                try:
                    this_cast = guess_type(this_str)
                except ValueError:
                    this_cast = this_str

            results.append(this_cast)
        return results

    def _read_column_names(self, f):
        return [x.split("(")[0] for x in f.readline().split()]




class AHFStatFile(HaloStatFile):
    _finder_offset_start = 1

    _column_translations = {'n_gas': translations.DefaultValue('n_gas', 0),
                            'n_star': translations.DefaultValue('n_star', 0),
                            'n_dm': translations.Function(lambda ngas, nstar, npart: npart - (ngas or 0) - (nstar or 0),
                                                          'n_gas', 'n_star', 'npart'),
                            'hostHalo': translations.Function(
                                lambda id: None if id==-1 else proxy_object.IncompleteProxyObjectFromFinderId(id, 'halo'),
                                'hostHalo')}

    def __init__(self, timestep_filename):
        super(AHFStatFile, self).__init__(timestep_filename)
        self._column_translations = copy.copy(self._column_translations)
        self._column_translations['childHalo'] = translations.Function(self._child_halo_entry, '#ID')

    @classmethod
    def filename(cls, timestep_filename):
        import glob
        file_list = glob.glob(timestep_filename+ '.z*.???.AHF_halos')

        # permit the AHF halos to be in a subfolder called "halos", for yt purposes
        # (where the yt tipsy reader can't cope with the AHF files being in the same folder)
        parts = timestep_filename.split("/")
        parts_with_halo = parts[:-1]+["halos"]+parts[-1:]
        filename_with_halo = "/".join(parts_with_halo)
        file_list+=glob.glob(filename_with_halo+'.z*.???.AHF_halos')

        if len(file_list)==0:
            return "CannotFindAHFHaloFilename"
        else:
            return file_list[0]

    def _calculate_children(self):
        # use hostHalo column to calculate virtual childHalo entries
        self._children_map = {}
        for c_id, f_id, host_f_id in self.iter_rows_raw("hostHalo"):
            if host_f_id!=-1:
                cmap = self._children_map.get(host_f_id, [])
                cmap.append(proxy_object.IncompleteProxyObjectFromFinderId(f_id,'halo'))
                self._children_map[host_f_id] = cmap

    def _calculate_children_if_required(self):
        if not hasattr(self, "_children_map"):
            self._calculate_children()

    def _child_halo_entry(self, this_id_raw):
        self._calculate_children_if_required()
        children = self._children_map.get(this_id_raw, [])
        return children

class RockstarStatFile(HaloStatFile):
    _column_translations = {'n_dm': translations.Rename('Np'),
                            'n_gas': translations.Value(0),
                            'n_star': translations.Value(0),
                            'npart': translations.Rename('Np')}

    @classmethod
    def filename(cls, timestep_filename):
        basename = os.path.basename(timestep_filename)
        dirname = os.path.dirname(timestep_filename)
        if basename.startswith("snapshot_"):
            timestep_id = int(basename[9:])
            return os.path.join(dirname, "out_%d.list"%timestep_id)
        else:
            return "CannotComputeRockstarFilename"

class AmigaIDLStatFile(HaloStatFile):

    _column_translations = {'n_dm': translations.Rename('N_dark'),
                            'n_gas': translations.Rename('N_gas'),
                            'n_star': translations.Rename("N_star"),
                            'npart': translations.Function(lambda ngas, nstar, ndark: ngas + nstar + ndark,
                                                           "N_dark", "N_gas", "N_star")}

    @classmethod
    def filename(cls, timestep_filename):
        return timestep_filename + '.amiga.stat'

    def iter_rows_raw(self, *args):
        """
        Yield the halo ID along with the values associated with each of the given arguments. The halo ID is output twice
        in order to be consistent with other stat file readers. In this case, the finder_offset that is normally output
        is just equal to the finder_id.

        :param args: strings for the column names
        :return: finder_id, finder_id, arg1, arg2, arg3, ... where finder_id is the halo's ID number read directly
        from the stat file and argN is the value associated with the Nth column name given as arguments.
        """

        for row in super().iter_rows_raw(*args):
            row[0] = row[1]  # sequential catalog index not right in this case; overwrite to match finder id
            yield row



from more_itertools import always_iterable
from ..util import proxy_object
from itertools import chain
from .pynbody import PynbodyInputHandler
import numpy as np
from ..log import logger
from .. import config

class RamsesCatalogueMixin:
    def create_bridge(self, f1, f2):
        import pynbody
        # Ensure that f1.dm and f2.dm are not garbage-collected
        self._f1dm = f1.dm
        self._f2dm = f2.dm

        return pynbody.bridge.OrderBridge(self._f1dm, self._f2dm, monotonic=False)

    def match_objects(self, ts1, ts2, halo_min, halo_max, dm_only=True, threshold=0.005,
                      object_typetag="halo", output_handler_for_ts2=None):
        import pynbody
        if not dm_only:
            logger.warn(
                "`match_objects` was called with dm_only=%s, but %s only supports DM-only"
                " catalogues at the moment. Falling back to DM-only.", dm_only, self.__class__.__name__
            )
            dm_only = True

        return super().match_objects(
            ts1,
            ts2,
            halo_min,
            halo_max,
            dm_only=dm_only,
            threshold=threshold,
            object_typetag=object_typetag,
            output_handler_for_ts2=output_handler_for_ts2,
            fuzzy_match_kwa={"use_family": pynbody.family.dm}
        )

class RamsesHOPInputHandler(RamsesCatalogueMixin, PynbodyInputHandler):
    """ Handling Ramses outputs with HOP halo finding (Eisenstein and Hut 1998)"""
    patterns = ["output_0????"]
    auxiliary_file_patterns = ["grp*.tag"]

    def _is_able_to_load(self, ts_extension):
        import pynbody
        filepath = self._extension_to_filename(ts_extension)
        try:
            f = pynbody.load(filepath)
            if self.quicker:
                logger.warn("Pynbody was able to load %r, but because 'quicker' flag is set we won't check whether it can also load the halo files", filepath)
            else:
                h = pynbody.halo.hop.HOPCatalogue(f)
            return True
        except (IOError, RuntimeError):
            return False



class RamsesAdaptaHOPInputHandler(RamsesCatalogueMixin, PynbodyInputHandler):
    """ Handling Ramses outputs with AdaptaHOP halo and subhalo finding """

    patterns = ["output_0????"]
    auxiliary_file_patterns = ["tree_bricks???"]

    _excluded_adaptahop_precalculated_properties = (
        "members", "timestep", "level", "host_id", "first_subhalo_id", "next_subhalo_id",
        "pos_x", "pos_y", "pos_z", "pos", "vel_x", "vel_y", "vel_z",
        "angular_momentum_x", "angular_momentum_y", "angular_momentum_z",
        "contaminated", "m_contam", "mtot_contam", "n_contam", "ntot_contam",
    )
    _included_adaptahop_additional_properties = (
        "parent", "child", "shrink_center", "bulk_velocity", "contamination_fraction"
    )

    def _is_able_to_load(self, ts_extension):
        import pynbody
        filepath = self._extension_to_filename(ts_extension)
        try:
            f = pynbody.load(filepath)
            if self.quicker:
                logger.warn("Pynbody was able to load %r, but because 'quicker' flag is set we won't check whether it can also load the halo files", filepath)
            else:
                h = f.halos()
                return isinstance(h, pynbody.halo.adaptahop.BaseAdaptaHOPCatalogue)
            return True
        except (IOError, RuntimeError):
            return False


    @staticmethod
    def _compute_contamination_fraction(adaptahop_halo):
        # Deal with the potential case that contamination fraction has not been computed through AdaptaHOP
        try:
            contamination_fraction = float(adaptahop_halo.properties['ntot_contam'] / adaptahop_halo.properties['ntot'])
        except KeyError:
            logger.warn("Ignoring import of contamination fraction which has not been stored on disk by AdaptaHOP")
            contamination_fraction = None
        return contamination_fraction

    def available_object_property_names_for_timestep(self, ts_extension, object_typetag):
        h = self._construct_halo_cat(ts_extension, object_typetag)

        halo_attributes = list(h._halo_attributes)
        if h._read_contamination:
            halo_attributes.extend(h._halo_attributes_contam)

        attrs = chain.from_iterable(tuple(always_iterable(attr)) for (attr, _len, _dtype) in halo_attributes)

        # Import all precalculated properties except conflicting ones with Tangos syntax
        property_list = [attr for attr in attrs if attr not in self._excluded_adaptahop_precalculated_properties]
        # Add additional properties that can be derived from adaptahop
        property_list += self._included_adaptahop_additional_properties
        return property_list

    @staticmethod
    def _resolve_units(value):
        import pynbody
        if (pynbody.units.is_unit(value)):
            return float(value)
        else:
            return value

    def _get_map_child_subhalos(self, ts_extension):
        h = self._construct_halo_cat(ts_extension, 'halo')
        halo_children = {}
        for halo_i in range(1, len(h)+1):  # AdaptaHOP catalogues start at 1
            halo_props = h[halo_i].properties

            if halo_props['host_id'] != halo_i: # If halo isn't its own host, it is a subhalo
                parent = halo_props['host_id']
                if parent not in halo_children:
                    halo_children[parent] = []
                halo_children[parent].append(halo_i)
        return halo_children

    def iterate_object_properties_for_timestep(self, ts_extension, object_typetag, property_names):
        h = self._construct_halo_cat(ts_extension, object_typetag)
        # For AdaptaHOP handler, we do not need to load the entire snaphshot to enumerate
        # properties in the halo catalogue. If pynbody supports this, ask it to do so.
        h._index_parent = False
        h.physical_units()  # make sure all units are physical before storing to database

        if "child" in property_names:
            # Construct the mapping between parent and subhalos
            map_child_parent = self._get_map_child_subhalos(ts_extension)

        for halo_i in range(1, len(h)+1):  # AdaptaHOP catalogues start at 1

            # Tangos expects data to have a finder offset, and a finder id following the stat file logic
            # I think these are irrelevant for AdaptaHOP catalogues which are derived directly from pynbody
            # Putting the finder ID twice seems to produce consistent results
            all_data = [halo_i, halo_i]

            adaptahop_halo = h[halo_i]
            precalculated_properties = adaptahop_halo.properties

            # Loop over all properties we wish to import
            for k in property_names:
                if k in self._included_adaptahop_additional_properties:
                    if k == "parent":
                        data = proxy_object.IncompleteProxyObjectFromFinderId(precalculated_properties['host_id'], 'halo')
                    elif k == "child":
                        # Determine whether halo has childs and create halo objects to it
                        try:
                            list_of_child = map_child_parent[halo_i]
                            data = [proxy_object.IncompleteProxyObjectFromFinderId(data_i, 'halo') for data_i in list_of_child]
                        except KeyError:
                            data = None
                    elif k == "shrink_center":
                        # Avoid naming confusions with already defined PynbodyProperties
                        data = (adaptahop_halo.properties['pos']).view(np.ndarray)
                    elif k == "bulk_velocity":
                        data = (adaptahop_halo.properties['vel']).view(np.ndarray)
                    elif k == "contamination_fraction":
                        data = self._compute_contamination_fraction(adaptahop_halo)
                    else:
                        raise NotImplementedError(
                            "Cannot handle property %s for halo catalogue %r" % (
                                k, self
                            )
                        )
                elif k in precalculated_properties:
                    data = precalculated_properties[k]
                    # Strip the unit as Tangos expects it to be a raw number
                    data = self._resolve_units(data)
                else:
                    data = None

                all_data.append(data)
            logger.info("Done with Halo %i" % halo_i)
            yield all_data

    def enumerate_objects(self, ts_extension, object_typetag="halo", min_halo_particles=config.min_halo_particles):
        if self._can_enumerate_objects_from_statfile(ts_extension, object_typetag):
            for X in self._enumerate_objects_from_statfile(ts_extension, object_typetag):
                yield X
        else:
            logger.warn("No halo statistics file found for timestep %r", ts_extension)

            try:
                h = self._construct_halo_cat(ts_extension, object_typetag)
                h._index_parent = False
            except:
                logger.warn("Unable to read %ss using pynbody; assuming step has none", object_typetag)
                return

            logger.warn(" => enumerating %ss directly using pynbody", object_typetag)
            istart = 1

            for i in range(istart, len(h)+istart):
                try:
                    hi = h[i]
                    if hi.properties["npart"] > min_halo_particles:
                        yield i, i, hi.properties["npart"], 0, 0
                except (ValueError, KeyError) as e:
                    pass



from __future__ import absolute_import
from sqlalchemy.orm import relationship



from .property import HaloProperty
from .link import HaloLink



def _initialise_halo_property_relationships():

    from ..halo import SimulationObjectBase
    from ..timestep import TimeStep

    SimulationObjectBase.properties = relationship(HaloProperty, cascade='all', lazy='dynamic',
                                   primaryjoin=(HaloProperty.halo_id == SimulationObjectBase.id) & (
                                       HaloProperty.deprecated == False),
                                   order_by=HaloProperty.id,
                                   uselist=True,overlaps='deprecated_properties')

    SimulationObjectBase.deprecated_properties = relationship(HaloProperty, cascade='all', lazy='dynamic',
                                              primaryjoin=(HaloProperty.halo_id == SimulationObjectBase.id) & (
                                                  HaloProperty.deprecated == True),
                                              order_by=HaloProperty.id,
                                              uselist=True,overlaps='properties')

    TimeStep.links_from = relationship(HaloLink, secondary=SimulationObjectBase.__table__,
                                       secondaryjoin=(
                                           HaloLink.halo_from_id == SimulationObjectBase.id),
                                       primaryjoin=(
                                           SimulationObjectBase.timestep_id == TimeStep.id),
                                       cascade='none', lazy='dynamic',
                                       order_by=HaloLink.id,
                                       viewonly=True)



    TimeStep.links_to = relationship(HaloLink, secondary=SimulationObjectBase.__table__,
                                     secondaryjoin=(
                                         HaloLink.halo_to_id == SimulationObjectBase.id),
                                     primaryjoin=(SimulationObjectBase.timestep_id == TimeStep.id),
                                     cascade='none', lazy='dynamic',
                                     order_by=HaloLink.id,
                                     viewonly=True)



    SimulationObjectBase.all_links = relationship(HaloLink, primaryjoin=(HaloLink.halo_from_id == SimulationObjectBase.id),
                                  viewonly=True, order_by=HaloLink.id)
    SimulationObjectBase.all_reverse_links = relationship(HaloLink, primaryjoin=(HaloLink.halo_to_id == SimulationObjectBase.id),
                                          viewonly=True, order_by=HaloLink.id)

_initialise_halo_property_relationships()


from .. import core

class TimestepObjectCache(object):
    """A temporary store for all objects in a timestep, allowing objects to be resolved without a further database query"""
    def __init__(self, timestep):
        """Query the database for all objects in this timestep

        :type timestep: core.TimeStep"""

        self.session = core.Session.object_session(timestep)
        self._timestep_id = timestep.id

    def _initialise_cache(self):
        all_objects = self.session.query(core.SimulationObjectBase).filter_by(timestep_id=self._timestep_id).all()

        self._map_finder_offset = {}
        self._map_finder_id = {}
        for obj in all_objects:
            typetag = obj.object_typetag_from_code(obj.object_typecode)
            if typetag not in self._map_finder_offset:
                self._map_finder_offset[typetag] = {}
            if typetag not in self._map_finder_id:
                self._map_finder_id[typetag] = {}
            self._map_finder_offset[typetag][obj.finder_offset] = obj
            self._map_finder_id[typetag][obj.finder_id] = obj

    def _ensure_cache(self):
        if not hasattr(self, "_map_finder_id") or not hasattr(self, "_map_finder_offset"):
            self._initialise_cache()

    def resolve_from_finder_offset(self, finder_offset, typetag):
        self._ensure_cache()
        try:
            return self._map_finder_offset[typetag][finder_offset]
        except KeyError:
            return None

    def resolve_from_finder_id(self, finder_id, typetag):
        self._ensure_cache()
        try:
            return self._map_finder_id[typetag][finder_id]
        except KeyError:
            return None


from . import PynbodyPropertyCalculation


class Contamination(PynbodyPropertyCalculation):
    """ Calculates the contamination of a zoomed halo by heavier, unzoomed particles.
        The current behaviour returns:
            1.0 if the halo contains only heavy particles
            X.X if the halo contains a mixture of heavy and light particles
            0.0 if the halo contains exclusively light particles
        Heavies are defined as any particle heavier than the deepest zoom level if multiple zooms are performed
    """

    names = "contamination_fraction"

    def calculate(self, halo, exist):
        loaded_data_min_dm_mass = halo.dm['mass'].min()

        # If mass resolution is not present as a simulation property,
        # ensure backward compatibility by setting the min mass to infinity
        import numpy as np
        import pynbody.array as array
        simulation_min_dm_mass = array.SimArray(self.get_simulation_property("approx_resolution_Msol", default=np.inf),
                                                units="Msol")

        if np.isclose(loaded_data_min_dm_mass, simulation_min_dm_mass, rtol=1e-1, atol=1):
            # Loaded data contains some light particles. Tolerance of this test (agreement to 10% or to 1 Msol)
            # is sufficiently tight to separate heavy and light resolution,
            # while being broad enough to capture numerical inaccuracies coming
            # from the simulation mass resolution.
            n_heavy = (halo.dm['mass'] > loaded_data_min_dm_mass).sum()
            return float(n_heavy) / len(halo.dm)
        else:
            # Loaded data contains only "heavy" particles, e.g. an unzoomed halo in server mode
            return 1.0



from __future__ import absolute_import

import re
import numpy as np
import six
from ..log import logger
import os


class BHLogData(object):
    """Class to load a Changa BH log files, either simname.BlackHoles or the (now deprecated) simname.shortened.orbit"""
    _cache = {}
    _n_cols = 0

    @classmethod
    def can_load(cls, filename):
        simname, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        try:
            return os.path.exists(cls.filename(simname))
        except (ValueError, TypeError):
            return False

    @classmethod
    def filename(cls, simname):
        raise ValueError("Unknown path to stat file")

    @classmethod
    def get_existing_or_new(cls, filename):
        name, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        obj = cls._cache.get(name, None)
        if obj is not None:
            return obj

        obj = cls(filename)
        cls._cache[name] = obj
        return obj

    def read_data(self, filename, sim):
        """
        generic function to read ascii columns from a given log file.
        :param filename: name of log file for black holes
        :param sim: loaded simulation object to connect the SimArray objects
        :return: simulation arrays for bh properties iord, time, step, mass, x, y, z, vx, vy, vz,
        current mdot, average mdot, change in mass, and scale factor
        """
        raise NotImplementedError

    def __init__(self, filename):
        import pynbody
        f = pynbody.load(filename)
        self.boxsize = float(f.properties['boxsize'].in_units('kpc', a=f.properties['a']))
        name, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        wrapped_ars = self.read_data(self.filename(name), f)
        iord, time, step, mass, x, y, z, vx, vy, vz, mdot, mdotmean, dMaccum, scalefac = wrapped_ars

        logger.info("Loaded a BH log with %d entries", len(time))

        iord[(iord<0)] = 2 * 2147483648 + iord[(iord < 0)]

        munits = f.infer_original_units("Msol")
        posunits = f.infer_original_units("kpc")
        velunits = f.infer_original_units("km s^-1")
        # potunits = velunits**2
        tunits = posunits / velunits
        # Eunits = munits*potunits
        # decorate with units

        x *= scalefac
        y *= scalefac
        z *= scalefac
        vx *= scalefac
        vy *= scalefac
        vz *= scalefac

        mass.units = munits
        x.units = y.units = z.units = posunits / pynbody.units.Unit('a')
        vx.units = vy.units = vz.units = velunits / pynbody.units.Unit('a')
        # pot.units = potunits
        time.units = tunits
        mdot.units = munits / tunits
        mdotmean.units = munits / tunits
        dMaccum.units = munits
        # E.units = Eunits

        x.convert_units('kpc')
        y.convert_units('kpc')
        z.convert_units('kpc')
        vx.convert_units('km s^-1')
        vy.convert_units('km s^-1')
        vz.convert_units('km s^-1')
        mdot.convert_units('Msol yr^-1')
        mdotmean.convert_units('Msol yr^-1')
        mass.convert_units("Msol")
        time.convert_units("Gyr")
        dMaccum.convert_units("Msol")
        # E.convert_units('erg')

        self.vars = {'bhid': iord, 'step': step, 'x': x, 'y': y, 'z': z,
                     'vx': vx, 'vy': vy, 'vz': vz, 'mdot': mdot, 'mdotmean': mdotmean,'mass': mass,
                     'time': time, 'dM': dMaccum}

    def get_at_stepnum(self, stepnum):
        mask = self.vars['step'] == stepnum
        return dict((k, v[mask]) for k, v in six.iteritems(self.vars))

    def get_at_stepnum_for_id(self, stepnum, bhid):
        vars = self.get_at_stepnum(stepnum)
        try:
            index = np.where(vars['bhid']==bhid)[0][0]
        except IndexError:
            raise ValueError("BH %d not found in step %d"%(bhid,stepnum))
        vars_this = dict([(k,v[index]) for k, v in six.iteritems(vars)])
        return vars_this

    def get_last_entry_for_id(self, bhid):
        mask = self.vars['bhid'] == bhid
        if mask.sum()==0:
            raise ValueError("No entries for BH %d"%bhid)
        ilast = np.argmax(self.vars['time'][mask])
        return dict((k, v[mask][ilast]) for k, v in six.iteritems(self.vars))

    def determine_merger_ratio(self, bhid_eaten, bhid_survivor):
        eaten_entries = self.get_last_entry_for_id(bhid_eaten)
        eaten_mass = eaten_entries['mass']
        survivor_entries = self.get_at_stepnum_for_id(eaten_entries['step'], bhid_survivor)
        survivor_premerger_mass = survivor_entries['mass']
        return eaten_mass/survivor_premerger_mass


    def get_for_named_snapshot(self, filename):
        name, stepnum = re.match("^(.*)\.(0[0-9]*)$", filename).groups()
        stepnum = int(stepnum)
        return self.get_at_stepnum(stepnum)

class BlackHolesLog(BHLogData):
    _n_cols = 18
    _col_types = [int, float, float, float, float, float,
                     float, float, float, float, float, float,
                     float, float, float, float, float, float]

    @classmethod
    def filename(cls, simname):
        return simname + '.BlackHoles'

    def read_data(self, filename, sim):
        import pynbody
        ars = [[] for i in range(self._n_cols)]
        for line in open(filename):
            line_split = line.split()
            for i in range(self._n_cols):
                ars[i].append(self._col_types[i](line_split[i]))

        wrapped_ars = [pynbody.array.SimArray(x) for x in ars]
        for w in wrapped_ars:
            w.sim = sim

        iord, time, step, mass, x, y, z, vx, vy, vz, pot, mdot, dM, dE, dt, dMaccum, dEaccum, scalefac \
            = wrapped_ars
        dt_out = np.zeros(len(iord))
        osort = np.argsort(step)

        unique_step, unique_ind, unique_step_inv = np.unique(step[osort], return_inverse=True, return_index=True)
        dt_ustep = time[osort[unique_ind[1:]]] - time[osort[unique_ind[:-1]]]
        dt_ustep = np.insert(dt_ustep, 0, dt_ustep[0])
        dt_out[osort] = dt_ustep[unique_step_inv]

        mdotmean = dMaccum/dt_out

        return iord, time, step, mass, x, y, z, vx, vy, vz, mdot, mdotmean, dMaccum, scalefac



class ShortenedOrbitLog(BHLogData):
    _n_cols = 15
    _col_types = [int, float, int, float, float, float,
                  float, float, float, float, float, float,
                  float, float, float]

    @classmethod
    def filename(cls, timestep_filename):
        return timestep_filename + '.shortened.orbit'

    def read_data(self, filename, sim):
        import pynbody
        ars = [[] for i in range(self._n_cols)]
        for line in open(filename):
            line_split = line.split()
            for i in range(self._n_cols):
                ars[i].append(self._col_types[i](line_split[i]))

        wrapped_ars = [pynbody.array.SimArray(x) for x in ars]
        for w in wrapped_ars:
            w.sim = sim

        iord, time, step, mass, x, y, z, vx, vy, vz, mdot, mdotmean, mdotsig, scalefac, dMaccum = wrapped_ars

        return iord, time, step, mass, x, y, z, vx, vy, vz, mdot, mdotmean, dMaccum, scalefac


from . import BuiltinFunction, StoredProperty
from .. import Link, MultiCalculation, ReturnInputHalos, FixedInput
from ...core import extraction_patterns
import numpy as np

@BuiltinFunction.register
def has_link(source_halos, link):
    return np.not_equal(link,None)

@has_link.set_initialisation
def link_exists_initialisation(input):
    input.set_extraction_pattern(
        extraction_patterns.HaloLinkTargetGetter())


@BuiltinFunction.register
def link(*args):
    """link(link_name) - get the first named link
    link(link_name, property_name, property_criterion) - get the named link where link_name.property_name is either maximum or minimum """
    raise RuntimeError("Internal error in link(). The customised function from the link_initialisation routine should be called.")

link.set_input_options(2, assert_class=FixedInput)
for i in range(20): # TODO: implement a way to set provide_proxy on all possible arguments, to remove arbitrary upper limit
    link.set_input_options(i, provide_proxy=True)

@link.set_initialisation
def link_initialisation(link_getter, property_getter=None, basis='max',*constraints):
    if isinstance(basis, FixedInput):
        basis = basis.proxy_value()
    if property_getter:
        if constraints:
            internal_getter = Link(link_getter, MultiCalculation(ReturnInputHalos(), property_getter,*constraints))
            internal_getter.set_constraints_columns(np.arange(len(constraints))+2)
        else:
            internal_getter = Link(link_getter, MultiCalculation(ReturnInputHalos(), property_getter))
        internal_getter.set_multi_selection_basis(basis, 1)
    else:
        internal_getter = Link(link_getter, ReturnInputHalos())
        internal_getter.set_multi_selection_basis('first')

    def custom_link(source_halos, *args):
        return internal_getter.values(source_halos)[0]

    return custom_link



from . import PynbodyPropertyCalculation

class Masses(PynbodyPropertyCalculation):
    names = "finder_mass"

    def calculate(self, halo, existing_properties):
        return halo['mass'].sum()


class MassBreakdown(PynbodyPropertyCalculation):
    names = "finder_dm_mass", "finder_star_mass", "finder_gas_mass"
    
    def calculate(self, halo, existing_properties):
        return halo.dm['mass'].sum(), halo.star['mass'].sum(), halo.gas['mass'].sum()



from . import PynbodyPropertyCalculation
from ...input_handlers import eagle
from ...log import logger
import numpy as np

class EagleBH(PynbodyPropertyCalculation):
    works_with_handler = eagle.EagleLikeInputHandler

    names = "BH_mass", "BH_mdot"

    def calculate(self, particle_data, halo_entry):
        return particle_data.bh['BH_Mass'].sum(), particle_data.bh['BH_Mdot'].sum()



"""Support for the directory structure used by Eagle-like runs"""

from .pynbody import PynbodyInputHandler
import os
import re
import weakref
import scipy.stats
import numpy as np
from .. import config
from ..util import proxy_object
from ..log import logger
from six.moves import range

_eagle_underlying_subfind_cache = weakref.WeakValueDictionary()

class EagleLikeInputHandler(PynbodyInputHandler):
    patterns = ["snapshot_???_z*"]
    
    @classmethod
    def _snap_id_from_snapdir_path(cls, path):
        match = re.match(".*snapshot_([0-9]{3}_z[0-9]{3}p[0-9]{3})/?", path)
        if match:
            return match.group(1)
        else:
            return None

    @classmethod
    def _pynbody_extension_from_ts_extension(cls, path):
        snap_id = cls._snap_id_from_snapdir_path(path)
        if snap_id:
            return os.path.join(path, "snap_%s" % snap_id)
        else:
            raise IOError("Cannot infer correct path to pass to pynbody")

    @classmethod
    def _pynbody_subfind_extension_from_ts_extension(cls, path):
        snap_id = cls._snap_id_from_snapdir_path(path)
        if snap_id:
            subfind_path = os.path.join(os.path.dirname(path), "particledata_%s/eagle_subfind_particles_%s" % (snap_id, snap_id))
            if os.path.exists(os.path.dirname(subfind_path)):
                return subfind_path
            else:
                return cls._pynbody_extension_from_ts_extension(path)
        else:
            raise IOError("Cannot infer correct subfind particledata path to pass to pynbody")

    def _extension_to_filename(self, ts_extension):
        return str(os.path.join(config.base, self.basename, self._pynbody_extension_from_ts_extension(ts_extension)))

    def _extension_to_halodata_filename(self, ts_extension):
        return str(os.path.join(config.base, self.basename, self._pynbody_subfind_extension_from_ts_extension(ts_extension)))

    def _is_able_to_load(self, ts_extension):
        from .pynbody import pynbody
        filepath = self._extension_to_filename(ts_extension)
        try:
            pynbody.load(filepath)
        except (IOError, RuntimeError):
            return False

        halofilepath = self._extension_to_halodata_filename(ts_extension)
        try:
            pynbody.load(halofilepath)
        except (IOError, RuntimeError):
            return False

        return True

    def _construct_halo_cat(self, ts_extension, object_typetag):
        from .pynbody import pynbody
        if object_typetag!= 'halo' and object_typetag!='group':
            raise ValueError("Unknown object type %r" % object_typetag)

        f = self.load_timestep(ts_extension)

        halofilepath = self._extension_to_halodata_filename(ts_extension)
        to_unpack = getattr(f, "_eagle_subfind_cache", None)

        if to_unpack is None:
            logger.debug("Eagle input handler constructing GrpCatalogues")
            if halofilepath==self._extension_to_filename(ts_extension):
                f_subfind = f
            else:
                f_subfind = pynbody.load(halofilepath)
            h_group = f_subfind.halos()

            # Eagle files have group and subgroup numbers, but tangos can only associate a single "finder id" with
            # each object, so we need to map these pairs to a sensible unique subgroup
            self._create_unique_subgroup_ids(f_subfind)

            h_halo = pynbody.halo.GrpCatalogue(f_subfind, 'TangosSubGroupNumber',
                                          ignore=f_subfind['TangosSubGroupNumber'].max())
            logger.debug("Eagle input handler found on-disk total of %d groups and %d subgroups" % (len(h_group), len(h_halo)))
            h_halo.precalculate()
            h_group.precalculate()
            f._eagle_subfind_cache = f_subfind, h_group, h_halo
        else:
            f_subfind, h_group, h_halo = to_unpack

        if object_typetag=='halo':
            return h_halo
        elif object_typetag=='group':
            return h_group
        else:
            assert False # should have been caught above

    def _create_unique_subgroup_ids(self, f_subfind):
        # see if we previously saved it:
        if 'TangosSubGroupNumber' in f_subfind.loadable_keys():
            logger.info("Found cached TangosSubGroupNumber within Eagle snapshot")
            return

        grp = f_subfind['GroupNumber']
        subgrp = f_subfind['SubGroupNumber']
        logger.info("Mapping Eagle-like group/subgroups to unique subgroups across snapshot")
        # group together all particles that are not in a subgroup. The subgroup files do not have any particles
        # that are not in a group.
        unique_subgrp_hash = (subgrp.astype(np.int64) + (grp.astype(np.int64) << 32)).view(type=np.ndarray)
        subgrp_max = int(subgrp.max())
        unique_subgrp_hash[subgrp == subgrp_max] = subgrp_max
        unique_subgrp_ordered = scipy.stats.rankdata(unique_subgrp_hash, 'dense')
        unique_subgrp_ordered[subgrp == subgrp_max] = subgrp_max
        f_subfind['TangosSubGroupNumber'] = unique_subgrp_ordered
        try:
            f_subfind['TangosSubGroupNumber'].write()
        except IOError:
            logger.info("Unable to cache TangosSubGroupNumber on disk")
        return subgrp_max

        
    def available_object_property_names_for_timestep(self, ts_extension, object_typetag):
        if object_typetag=='halo':
            return ['parent', 'original_subgroup_number']
        elif object_typetag=='group':
            return ['child']
        else:
            raise ValueError("Unknown object type %r"%object_typetag)

    @staticmethod
    def _second_largest(array):
        max_ = np.max(array)
        second_max = np.max(array[array!=max_])
        return second_max

    def iterate_object_properties_for_timestep(self, ts_extension, object_typetag, property_names):
        from .pynbody import pynbody
        if object_typetag not in ("halo","group"):
            raise ValueError("Unknown object type tag %r"%object_typetag)
        
        halofilepath = self._extension_to_halodata_filename(ts_extension)
        f_subfind = pynbody.load(halofilepath)
        self._create_unique_subgroup_ids(f_subfind)

        logger.info("Calculating child-parent relationships")
        children, unique_indices = np.unique(f_subfind['TangosSubGroupNumber'],return_index=True)
        parents = f_subfind['GroupNumber'][unique_indices]
        original = f_subfind['SubGroupNumber'][unique_indices]

        logger.info("Done; enumerating properties")
        if object_typetag=='halo':
            num_children = self._second_largest(children)
            for i in range(num_children):
                row, = np.where(children==i)
                if len(row)==0:
                    props = {}
                else:
                    props = {'parent': proxy_object.IncompleteProxyObjectFromFinderId(parents[row[0]],'group'),
                             'original_subgroup_number': original[row[0]]}
                yield [i]+[props.get(n, None) for n in property_names]
        else:
            # must be group
            num_parents = self._second_largest(parents)
            for i in range(num_parents):
                row, = np.where(parents==i)
                children_proxies = [proxy_object.IncompleteProxyObjectFromFinderId(child, 'halo') for child in children[row]]
                if len(children_proxies)==0:
                    props = {}
                else:
                    props = {'child': children_proxies}
                yield [i]+[props.get(n, None) for n in property_names]




#!/usr/bin/env python2.7

from __future__ import absolute_import
from __future__ import print_function
import tangos as db
from tangos.util.check_deleted import check_deleted
import tangos.core.dictionary
import tangos.core.halo
import tangos.core.halo_data
import tangos.core.timestep
import tangos.core.tracking
import tangos.parallel_tasks as parallel_tasks
import tangos.parallel_tasks.database
import tangos.tracking
from tangos.input_handlers.changa_bh import ShortenedOrbitLog, BlackHolesLog
from sqlalchemy.orm import Session
from tangos.log import logger
import sys
import numpy as np
import gc
import argparse
import glob
import pynbody
import six
from six.moves import zip

def get_parser_object():
    parser = argparse.ArgumentParser()
    tangos.core.supplement_argparser(parser)
    parser.add_argument('--sims', '--for', action='store', nargs=1,
                        metavar='simulation_name',
                        help='Specify a simulation to run on')
    parser.add_argument('--link-only', action='store_true',
                        help='Do not look for any new black holes; only link already-known black holes to new timesteps or halos')
    parser.add_argument('--backwards', action='store_true',
                        help="Scan through timesteps from low to high redshift (default is high to low)")
    return parser

def generate_missing_bh_objects(bh_iord, f, existing_obj_num):
    halo = []
    for bhi in bh_iord:
        if bhi not in existing_obj_num:
            halo.append(tangos.core.halo.BH(f, int(bhi)))

    return halo

def collect_bh_trackers(bh_iord,sim,existing_trackers):
    track = []
    for bhi in bh_iord:
        bhi = int(bhi)
        et = np.where(existing_trackers == bhi)[0]
        if len(et)==0:
            tx = tangos.core.tracking.TrackData(sim, bhi)
            tx.particles = [bhi]
            tx.use_iord = True
            track.append(tx)
    return track


def bh_halo_assign(f_pb):
    f_pbh = f_pb.halos()
    bh_cen_halos=None
    bh_halos = None
    if type(f_pbh) == pynbody.halo.RockstarIntermediateCatalogue:
        bh_cen_halos = f_pbh.get_group_array(family = 'BH')
    if type(f_pbh) == pynbody.halo.AHFCatalogue:
        bh_cen_halos = f_pbh.get_group_array(top_level=False, family = 'bh')
        bh_halos = f_pbh.get_group_array(top_level=True, family='bh')
    if type(f_pbh) != pynbody.halo.AHFCatalogue and type(f_pbh) != pynbody.halo.RockstarIntermediateCatalogue:
        f_pb['gp'] = f_pbh.get_group_array()
        bh_cen_halos = f_pb.star['gp'][np.where(f_pb.star['tform']<0)[0]]
    del f_pbh
    gc.collect()
    return bh_cen_halos, bh_halos


def scan_for_BHs(files, session):
    for timestep in parallel_tasks.distributed(files):
        logger.info("Processing %s", timestep)

        try:
            timestep_particle_data = timestep.load()
        except:
            logger.warning("File not found - continuing")
            continue

        if len(timestep_particle_data.star) < 1:
            logger.warning("No stars - continuing")
            continue

        timestep_particle_data.physical_units()

        logger.info("Gathering existing BH halo information from database for step %r", timestep)

        bhobjs = timestep.bhs.all()
        existing_bh_nums = [x.halo_number for x in bhobjs]

        logger.info("...found %d existing BHs", len(existing_bh_nums))

        logger.info("Gathering BH info from simulation for step %r", timestep)
        bh_iord_this_timestep = timestep_particle_data.star['iord'][
            np.where(timestep_particle_data.star['tform'] < 0)[0]]
        bh_mass_this_timestep = timestep_particle_data.star['mass'][
            np.where(timestep_particle_data.star['tform'] < 0)[0]]

        logger.info("Found %d black holes for %r", len(bh_iord_this_timestep), timestep)

        logger.info("Updating BH trackdata and BH objects using on-disk information from %r", timestep)

        add_missing_trackdata_and_BH_objects(timestep, bh_iord_this_timestep, existing_bh_nums, session)
        session.expire_all()

        logger.info("Calculating halo associations for BHs in timestep %r", timestep)
        bh_cen_halos, bh_halos = bh_halo_assign(timestep_particle_data)

        # re-order our information so that links refer to BHs in descending order of mass
        bh_order_by_mass = np.argsort(bh_mass_this_timestep)[::-1]
        bh_iord_this_timestep = bh_iord_this_timestep[bh_order_by_mass]
        if bh_halos is not None:
            bh_halos = bh_halos[bh_order_by_mass]
        if bh_cen_halos is not None:
            bh_cen_halos = bh_cen_halos[bh_order_by_mass]

        logger.info("Freeing the timestep particle data")
        with check_deleted(timestep_particle_data):
            del (timestep_particle_data)

        if bh_halos is not None:
            assign_bh_to_halos(bh_halos, bh_iord_this_timestep, timestep, "BH")
        if bh_cen_halos is not None:
            assign_bh_to_halos(bh_cen_halos, bh_iord_this_timestep, timestep, "BH_central", "host_halo")


def assign_bh_to_halos(bh_halo_assignment, bh_iord, timestep, linkname, hostname=None):
    session = Session.object_session(timestep)
    linkname_dict_id = tangos.core.dictionary.get_or_create_dictionary_item(session, linkname)
    if hostname is not None:
        host_dict_id = tangos.core.dictionary.get_or_create_dictionary_item(session, hostname)
    else:
        host_dict_id = None

    logger.info("Gathering %s links for step %r", linkname, timestep)

    links, link_id_from, link_id_to = db.tracking.get_tracker_links(session, linkname_dict_id)
    halos = timestep.halos.filter_by(object_typecode=0).all()

    halo_nums = [h.halo_number for h in halos]
    halo_catind = [h.finder_offset for h in halos]
    halo_ids = np.array([h.id for h in halos])

    logger.info("Gathering bh halo information for %r", timestep)
    with parallel_tasks.lock.SharedLock("bh"):
        bh_database_object, existing_bh_nums, bhobj_ids = get_bh_objs_numbers_and_dbids(timestep)

    bh_links = []

    with session.no_autoflush:
        for bhi, haloi in zip(bh_iord, bh_halo_assignment):
            haloi = int(haloi)
            bhi = int(bhi)
            if haloi not in halo_catind:
                logger.warn("Skipping BH in halo %d as no corresponding halo found in the database", haloi)
                continue
            if bhi not in existing_bh_nums:
                logger.warn("Can't find the database object for BH %d", bhi)
                print(bhi)
                print(existing_bh_nums)
                continue

            bh_index_in_list = existing_bh_nums.index(bhi)
            halo_index_in_list = halo_catind.index(haloi)
            bh_obj = bh_database_object[bh_index_in_list]
            halo_obj = halos[halo_index_in_list]

            num_existing_links = ((link_id_from == halo_ids[halo_index_in_list]) & (link_id_to == bhobj_ids[bh_index_in_list])).sum()
            if num_existing_links==0:
                bh_links.append(tangos.core.halo_data.HaloLink(halo_obj, bh_obj, linkname_dict_id))
                if host_dict_id is not None:
                    bh_links.append(tangos.core.halo_data.HaloLink(bh_obj, halo_obj, host_dict_id))

    logger.info("Committing %d %s links for step %r...", len(bh_links), linkname, timestep)
    with parallel_tasks.ExclusiveLock("bh"):
        session.add_all(bh_links)
        session.commit()
    logger.info("...done")


def get_bh_objs_numbers_and_dbids(timestep):
    bh_database_object = timestep.bhs.all()
    existing_bh_nums = [x.halo_number for x in bh_database_object]
    bhobj_ids = np.array([x.id for x in bh_database_object])
    return bh_database_object, existing_bh_nums, bhobj_ids


def add_missing_trackdata_and_BH_objects(timestep, this_step_bh_iords, existing_bhobj_iords, session):
    with parallel_tasks.ExclusiveLock("bh"):
        track, track_nums = db.tracking.get_trackers(timestep.simulation)
        with session.no_autoflush:
            tracker_to_add = collect_bh_trackers(this_step_bh_iords, timestep.simulation, track_nums)
            halo_to_add = generate_missing_bh_objects(this_step_bh_iords, timestep, existing_bhobj_iords)

        session.add_all(tracker_to_add)
        session.add_all(halo_to_add)
        session.commit()
    logger.info("Committed %d new trackdata and %d new BH objects for %r", len(tracker_to_add), len(halo_to_add), timestep)



def resolve_multiple_mergers(bh_map):
    for k,v in six.iteritems(bh_map):
        if v[0] in bh_map:
            old_target = v[0]
            old_weight = v[1]
            bh_map[k] = bh_map[old_target][0],v[1]*bh_map[old_target][1]
            logger.info("Multi-merger detected; reassigning %d->%d (old) %d (new)",k,old_target,bh_map[k][0])
            resolve_multiple_mergers(bh_map)
            return

def generate_halolinks(session, fname, pairs):
    for ts1, ts2 in parallel_tasks.distributed(pairs):
        bh_log = None
        if BlackHolesLog.can_load(ts2.filename):
            bh_log = BlackHolesLog(ts2.filename)
        elif ShortenedOrbitLog.can_load(ts2.filename):
            bh_log = ShortenedOrbitLog(ts2.filename)
        if bh_log is None:
            logger.error("Warning! No orbit file found!")
        links = []
        mergers_links = []
        bh_map = {}
        logger.info("Gathering BH tracking information for steps %r and %r", ts1, ts2)
        with parallel_tasks.ExclusiveLock("bh"):
            dict_obj = db.core.get_or_create_dictionary_item(session, "tracker")
            dict_obj_next = db.core.get_or_create_dictionary_item(session, "BH_merger_next")
            dict_obj_prev = db.core.get_or_create_dictionary_item(session, "BH_merger_prev")

        track_links_n, idf_n, idt_n = db.tracking.get_tracker_links(session, dict_obj_next)
        bh_objects_1, nums1, id1 = get_bh_objs_numbers_and_dbids(ts1)
        bh_objects_2, nums2, id2 = get_bh_objs_numbers_and_dbids(ts2)
        tracker_links, idf, idt = db.tracking.get_tracker_links(session,dict_obj)

        idf_n = np.array(idf_n)
        idt_n = np.array(idt_n)

        if len(nums1) == 0 or len(nums2) == 0:
            logger.info("No BHs found in either step %r or %r... moving on", ts1, ts2)
            continue

        logger.info("Generating BH tracker links between steps %r and %r", ts1, ts2)
        o1 = np.where(np.in1d(nums1,nums2))[0]
        o2 = np.where(np.in1d(nums2,nums1))[0]
        if len(o1) == 0 or len(o2) == 0:
            continue
        with session.no_autoflush:
            for ii, jj in zip(o1,o2):
                if nums1[ii] != nums2[jj]:
                    raise RuntimeError("BH iords are mismatched")
                exists = np.where((idf==id1[ii])&(idt==id2[jj]))[0]
                if len(exists) == 0:
                    links.append(tangos.core.halo_data.HaloLink(bh_objects_1[ii],bh_objects_2[jj],dict_obj,1.0))
                    links.append(tangos.core.halo_data.HaloLink(bh_objects_2[jj],bh_objects_1[ii],dict_obj,1.0))
        logger.info("Generated %d tracker links between steps %r and %r", len(links), ts1, ts2)

        logger.info("Generating BH Merger information for steps %r and %r", ts1, ts2)
        for l in open(fname[0]):
            l_split = l.split()
            t = float(l_split[6])
            bh_dest_id = int(l_split[0])
            bh_src_id = int(l_split[1])
            ratio = float(l_split[4])

            # ratios in merger file are ambiguous (since major progenitor may be "source" rather than "destination")
            # re-establish using the log file:
            try:
                ratio = bh_log.determine_merger_ratio(bh_src_id, bh_dest_id)
            except (ValueError, AttributeError) as e:
                logger.debug("Could not calculate merger ratio for %d->%d from the BH log; assuming the .BHmergers-asserted value is accurate",
                            bh_src_id, bh_dest_id)

            if t>ts1.time_gyr and t<=ts2.time_gyr:
                bh_map[bh_src_id] = (bh_dest_id, ratio)

        resolve_multiple_mergers(bh_map)
        logger.info("Gathering BH merger links for steps %r and %r", ts1, ts2)
        with session.no_autoflush:
            for src,(dest,ratio) in six.iteritems(bh_map):
                if src not in nums1 or dest not in nums2:
                    logger.warn("Can't link BH %r -> %r; missing BH objects in database",src,dest)
                    continue
                bh_src_before = bh_objects_1[nums1.index(src)]
                bh_dest_after = bh_objects_2[nums2.index(dest)]

                if ((idf_n==bh_src_before.id)&(idt_n==bh_dest_after.id)).sum()==0:
                    mergers_links.append(tangos.core.halo_data.HaloLink(bh_src_before,bh_dest_after,dict_obj_next,1.0))
                    mergers_links.append(tangos.core.halo_data.HaloLink(bh_dest_after,bh_src_before,dict_obj_prev,ratio))

        logger.info("Generated %d BH merger links for steps %r and %r", len(mergers_links), ts1, ts2)

        with parallel_tasks.ExclusiveLock("bh"):
            logger.info("Committing total %d BH links for steps %r and %r", len(mergers_links)+len(links), ts1, ts2)
            session.add_all(links)
            session.add_all(mergers_links)
            session.commit()
            logger.info("Finished committing BH links for steps %r and %r", ts1, ts2)

def timelink_bh(sims, session):
    query = db.sim_query_from_name_list(sims, session)
    for sim in query.all():
        pairs = list(zip(sim.timesteps[:-1],sim.timesteps[1:]))
        fname = glob.glob(db.config.base+"/"+sim.basename+"/*.BHmergers")
        if len(fname)==0:
            logger.error("No merger file for "+sim.basename)
            return
        elif len(fname)>1:
            logger.error("Can't work out which is the merger file for "+sim.basename)
            logger.error("Found: %s",fname)
            return
        with session.no_autoflush:
            generate_halolinks(session, fname, pairs)



def run():
    session = db.core.get_default_session()
    args = get_parser_object().parse_args(sys.argv[1:])

    query = db.sim_query_from_name_list(args.sims, session)

    files = db.core.get_default_session().query(tangos.core.timestep.TimeStep).filter(
        tangos.core.timestep.TimeStep.simulation_id.in_([q.id for q in query.all()])). \
        order_by(tangos.core.timestep.TimeStep.time_gyr).all()

    if args.backwards:
        files = files[::-1]


    parallel_tasks.database.synchronize_creator_object(session)

    if not args.link_only:
        scan_for_BHs(files, session)

    timelink_bh(args.sims, session)


def main():
    parallel_tasks.launch(run)







from .. import core, all_simulations, get_simulation, get_timestep, get_object
from ..log import logger
import six
import numpy.testing as npt
from sqlalchemy.orm import joinedload, object_session, undefer, Session
import numpy as np

class TangosDbDiff(object):
    """Class to compare two databases, used by the tangos diff command line tool"""

    def __init__(self, uri1, uri2, max_objects=10, ignore_keys=[]):
        if isinstance(uri1, Session):
            self.session1 = uri1
            uri1 = str(uri1.connection().engine)
        else:
            core.init_db(uri1)
            self.session1 = core.get_default_session()


        if isinstance(uri2, Session):
            self.session2 = uri2
            uri2 = str(uri2.connection().engine)
        else:
            core.init_db(uri2)
            self.session2 = core.get_default_session()

        self.test_simulations = True
        self.test_timesteps = True
        self.test_objects = True
        self.max_objects = max_objects
        self.test_properties = True
        self.ignore_keys = ignore_keys
        self.failed = False

        logger.info("Database 1 = %s",uri1)
        logger.info("Database 2 = %s",uri2)
        logger.info("For each timestep will check data in first %d objects of each type",max_objects)
        if len(ignore_keys)>0:
            logger.info("The value of properties with the names %s will be ignored.",", ".join(ignore_keys))
            logger.info("Only the existence and type of these properties will be checked.")

    def fail(self, message, *args):
        logger.error(message, *args)
        self.failed = True

    def compare(self):
        sims1 = set([s.basename for s in all_simulations(self.session1)])
        sims2 = set([s.basename for s in all_simulations(self.session2)])

        if len(sims1-sims2)>0:
            self.fail("Database 1 has additional simulations: %s", list(sims1-sims2))
        if len(sims2-sims1)>0:
            self.fail("Database 2 has additional simulations: %s", list(sims2-sims1))

        if self.test_simulations:
            for sim in sims1.intersection(sims2):
                self.compare_simulation(sim)

    def compare_simulation(self, sim):
        logger.info("Comparing simulation %s",sim)
        sim1 = get_simulation(sim, self.session1)
        sim2 = get_simulation(sim, self.session2)

        ts1 = set([ts.extension for ts in sim1.timesteps])
        ts2 = set([ts.extension for ts in sim2.timesteps])

        self._check_same_set(ts1, ts2, str(sim), 'timesteps')

        if self.test_timesteps:
            for ts in ts1.intersection(ts2):
                self.compare_timestep(sim+"/"+ts)

    def compare_timestep(self, ts):
        logger.info("Comparing timestep %s", ts)
        ts1 = get_timestep(str(ts), self.session1)
        ts2 = get_timestep(str(ts), self.session2)

        obj_filter = (core.SimulationObjectBase.halo_number < self.max_objects) | (core.SimulationObjectBase.object_typecode == core.SimulationObjectBase.object_typecode_from_tag('bh'))

        objects1 = dict([(o.path, o) for o in ts1.objects.filter(obj_filter).all()])
        objects2 = dict([(o.path, o) for o in ts2.objects.filter(obj_filter).all()])
        self._check_same_set(objects1.keys(), objects2.keys(), str(ts), 'objects')


        if self.test_objects:
            keys = set(objects1.keys()).intersection(objects2.keys())
            for k in keys:
                self._compare_objects(objects1[k], objects2[k])


    def _joined_links_load(self, parent):
        return object_session(parent).query(core.HaloLink).with_parent(parent, "all_links").\
            options(joinedload("halo_to"))

    def _joined_properties_load(self, parent):
        return object_session(parent).query(core.HaloProperty).with_parent(parent, "all_properties").\
            options(undefer("*")).options(joinedload("name"))

    def compare_object(self, obj):
        obj1 = get_object(obj, self.session1)
        obj2 = get_object(obj, self.session2)
        self._compare_objects(obj1, obj2)

    def _compare_objects(self, obj1, obj2):
        properties1 = dict([(prop.name.text, prop.data_raw) for prop in self._joined_properties_load(obj1)])
        properties2 = dict([(prop.name.text, prop.data_raw) for prop in self._joined_properties_load(obj2)])

        self._check_dict_same(properties1, properties2, obj1.path)

        links1 = dict([(link.relation.text+"->"+link.halo_to.path, link.weight) for link in self._joined_links_load(obj1)])
        links2 = dict([(link.relation.text+"->"+link.halo_to.path, link.weight) for link in self._joined_links_load(obj2)])

        self._check_dict_same(links1, links2, obj1.path, 'links')

    def _check_dict_same(self, properties1, properties2, name_of_object, name_of_things='properties'):
        prop1_names = set(properties1.keys())
        prop2_names = set(properties2.keys())
        self._check_same_set(prop1_names, prop2_names, name_of_object, name_of_things)
        self._check_almost_equal(properties1, properties2, name_of_object)

    def _check_same_set(self, objects1, objects2, name_of_object, name_of_things):
        objects1 = set(objects1)
        objects2 = set(objects2)
        self._check_setdiff_null("Database 1", objects1, objects2, name_of_object, name_of_things)
        self._check_setdiff_null("Database 2",objects2, objects1, name_of_object, name_of_things)

    def _check_setdiff_null(self, name_of_db_1, objects1, objects2,  name_of_object, name_of_things):
        if len(objects1 - objects2) > 0:
            self.fail("When comparing %s, %s has %d additional %s", name_of_object, name_of_db_1,
                      len(objects1 - objects2), name_of_things)
            self._print_objects(objects1 - objects2)

    def _print_objects(self, objects):
        for o in objects:
            logger.info("  %s", o)

    def _check_almost_equal(self, dict1, dict2, name_of_object):
        for d in dict1.keys():
            if d in dict2:
                v1 = dict1[d]
                v2 = dict2[d]
                if d in self.ignore_keys:
                    self._check_properties_compatible(v1, v2, d, name_of_object)
                else:
                    self._check_properties_equal(v1, v2, d, name_of_object)

    def _check_properties_equal(self, v1, v2, d, name_of_object):
        try:
            self._assert_equal(v1, v2)
        except AssertionError as e:
            self.fail("When comparing %s, value mismatch for key %s", name_of_object, d)
            self.fail("%s", e)

    def _check_properties_compatible(self, v1, v2, name_of_property, name_of_object):
        if type(v1) != type(v2):
            self.fail("When comparing %s, types of key %s differ", name_of_object, name_of_property)
        else:
            if isinstance(v1, np.ndarray):
                if v1.shape != v2.shape:
                    self.fail("When comparing %s, array shapes of key %s differ", name_of_object, name_of_property)

    def _assert_equal(self, v1, v2):
        if v1 is None or v2 is None:
            assert v1 is None, "Value 2 is None but Value 1 is not None"
            assert v2 is None, "Value 1 is None but Value 2 is not None"
        elif isinstance(v1, six.string_types):
            assert v1 == v2
        elif isinstance(v1, tuple):
            for e1, e2 in zip(v1,v2):
                self._assert_equal(e1, e2)
        else:
            npt.assert_allclose(v1, v2, rtol=1e-3, atol=1e-3)



from __future__ import absolute_import
from __future__ import print_function
import tangos.core.halo
from tangos.input_handlers.changa_bh import ShortenedOrbitLog, BlackHolesLog
from . import PynbodyPropertyCalculation
from .. import LivePropertyCalculation, TimeChunkedProperty
import numpy as np
import scipy, scipy.interpolate


class BH(PynbodyPropertyCalculation):

    names = "BH_mdot", "BH_mdot_ave", "BH_central_offset", "BH_central_distance", "BH_mass"
    requires_particle_data = True


    def requires_property(self):
        return ['host_halo']

    @classmethod
    def no_proxies(self):
        return True

    def preloop(self, f, db_timestep):
        if BlackHolesLog.can_load(db_timestep.filename):
            self.log = BlackHolesLog.get_existing_or_new(db_timestep.filename)
        elif ShortenedOrbitLog.can_load(db_timestep.filename):
            self.log = ShortenedOrbitLog.get_existing_or_new(db_timestep.filename)
        else:
            raise RuntimeError("cannot find recognizable log file")
        self.filename = db_timestep.filename
        print(self.log)

    def calculate(self, halo, properties):
        if not isinstance(properties, tangos.core.halo.Halo):
            raise RuntimeError("No proxies, please")
        boxsize = self.log.boxsize

        vars = self.log.get_for_named_snapshot(self.filename)

        mask = vars['bhid'] == properties.halo_number
        if (mask.sum() == 0):
            raise RuntimeError("Can't find BH in .orbit file")

        # work out who's the main halo
        # main_halo = None
        # for i in properties.reverse_links:
        #    if i.relation.text.startswith("BH"):
        #        main_halo = i.halo_from
        #        break
        # if main_halo is None:
        #    raise RuntimeError("Can't relate BH to its parent halo")

        try:
            main_halo = properties['host_halo']
        except KeyError:
            main_halo = None

        if main_halo is None:
            main_halo_ssc = None
        else:
            try:
                main_halo_ssc = main_halo['shrink_center']
            except KeyError:
                main_halo_ssc = None

        entry = np.where(mask)[0]

        print("target entry is", entry)
        final = {}
        for t in 'x', 'y', 'z', 'vx', 'vy', 'vz', 'mdot', 'mass', 'mdotmean':
            final[t] = float(vars[t][entry])

        if main_halo_ssc is None:
            offset = np.array((0, 0, 0))
        else:
            offset = np.array((final['x'], final['y'], final['z'])) - main_halo_ssc
            bad, = np.where(np.abs(offset) > boxsize / 2.)
            offset[bad] = -1.0 * (offset[bad] / np.abs(offset[bad])) * np.abs(boxsize - np.abs(offset[bad]))

        return final['mdot'], final['mdotmean'], offset, np.linalg.norm(offset), final['mass']


class BHAccHistogram(TimeChunkedProperty):

    requires_particle_data = True

    @classmethod
    def name(self):
        return "BH_mdot_histogram"

    def requires_property(self):
        return []

    def preloop(self, f, db_timestep):
        if BlackHolesLog.can_load(db_timestep.filename):
            self.log = BlackHolesLog.get_existing_or_new(db_timestep.filename)
        elif ShortenedOrbitLog.can_load(db_timestep.filename):
            self.log = ShortenedOrbitLog.get_existing_or_new(db_timestep.filename)
        else:
            raise RuntimeError("cannot find recognizable log file")

    @classmethod
    def no_proxies(self):
        return True

    def plot_xlabel(self):
        return "t/Gyr"

    def plot_ylabel(self):
        return r"$\dot{M}/M_{\odot}\,yr^{-1}$"

    def calculate(self, halo, properties):

        halo = halo.s

        if len(halo) != 1:
            raise RuntimeError("Not a BH!")

        if halo['tform'][0] > 0:
            raise RuntimeError("Not a BH!")

        mask = self.log.vars['bhid'] == halo['iord']
        if (mask.sum() == 0):
            raise RuntimeError("Can't find BH in .orbit file")

        t_orbit = self.log.vars['time'][mask]
        Mdot_orbit = self.log.vars['mdotmean'][mask]
        order = np.argsort(t_orbit)

        t_max = properties.timestep.time_gyr

        grid_tmax_Gyr = 20.0
        nbins = int(grid_tmax_Gyr/self.pixel_delta_t_Gyr)
        t_grid = np.linspace(0, grid_tmax_Gyr, nbins)

        Mdot_grid = scipy.interpolate.interp1d(t_orbit[order], Mdot_orbit[order], bounds_error=False)(t_grid)

        return Mdot_grid[self.store_slice(t_max)]


class BHAccHistogramMerged(PynbodyPropertyCalculation):
    names = "BH_mdot_histogram_all"

    @classmethod
    def no_proxies(self):
        return True

    def requires_property(self):
        return ["BH_mdot_histogram"]

    @classmethod
    def accumulate_on_mergers(cls, array, bh):
        while bh is not None:
            if "BH_merger" in list(bh.keys()):
                for targ_bh in bh.get_data("BH_merger", always_return_array=True):
                    if targ_bh.timestep.time_gyr < bh.timestep.time_gyr:
                        try:
                            accum = targ_bh["BH_mdot_histogram"]
                            array[:len(accum)][accum == accum] += accum[accum == accum]
                        except KeyError:
                            pass
                        cls.accumulate_on_mergers(array, targ_bh)

            bh = bh.previous

    def calculate(self, simdata, bh):
        mdot = bh['BH_mdot_histogram']
        self.accumulate_on_mergers(mdot, bh)
        return mdot


class BHGal(LivePropertyCalculation):
    def __init__(self, simulation=None, choose='BH_mass', minmax='max', bhtype='BH_central'):
        super(BHGal, self).__init__(simulation)
        self._maxmin = minmax
        self._choicep = choose
        self._bhtype = bhtype

    names = 'bh'

    def requires_property(self):
        return self._bhtype, self._bhtype+"."+self._choicep

    def live_calculate(self, halo, *args):
        if halo.object_typecode != 0:
            return None

        if self._bhtype not in list(halo.keys()):
            return None

        if type(halo[self._bhtype]) is list:
            chp = [bh[self._choicep] for bh in halo[self._bhtype] if self._choicep in bh]
            if len(chp) == 0:
                return None
            target = None
            if self._maxmin == 'min':
                target = np.argmin(chp)
            if self._maxmin == 'max':
                target = np.argmax(chp)
            if target is None:
                return None
            return halo[self._bhtype][target]
        else:
            return halo[self._bhtype]

class BHCentral(BHGal):
    names = 'bhcenter'

    def __init__(self, simulation=None):
        super(BHCentral, self).__init__(simulation, 'BH_central_distance', 'min', 'BH_central')



"""Specialised input handler for the caterpillar project
http://www.caterpillarproject.org"""

from __future__ import absolute_import

import re
from .. import config
from .pynbody import PynbodyInputHandler
from . import halo_stat_files
import os.path

class CaterpillarInputHandler(PynbodyInputHandler):
    patterns = ["snapdir_???"]

    @classmethod
    def _snap_id_from_snapdir_path(cls, path):
        match = re.match(".*snapdir_([0-9]{3})/?", path)
        if match:
            return int(match.group(1))
        else:
            return None

    @classmethod
    def _pynbody_path_from_snapdir_path(cls, path):
        snap_id = cls._snap_id_from_snapdir_path(path)
        if snap_id:
            return os.path.join(path, "snap_%.3d" % snap_id)
        else:
            raise IOError("Cannot infer correct path to pass to pynbody")

    @classmethod
    def _rockstar_path_from_snapdir_path(cls, path):
        snap_id = cls._snap_id_from_snapdir_path(path)
        if snap_id:
            return os.path.join(os.path.dirname(os.path.dirname(path)),
                                "halos", "halos_%d"%snap_id)
        else:
            raise IOError("Cannot infer path of halos")

    def _extension_to_filename(self, ts_extension):
        return str(os.path.join(config.base, self.basename, self._pynbody_path_from_snapdir_path(ts_extension)))

    def _is_able_to_load(self, filepath):
        try:
            import pynbody
            f = pynbody.load(self._pynbody_path_from_snapdir_path(filepath))
            h = pynbody.halo.RockstarCatalogue(f, pathname=self._rockstar_path_from_snapdir_path(filepath),
                                               format_revision='caterpillar')
            return True
        except (IOError, RuntimeError):
            return False

class CaterpillarRockstarStatFile(halo_stat_files.RockstarStatFile):

    @classmethod
    def filename(cls, timestep_filename):
        # following logic is horrible hack: should be solved by requiring consistency between filename passed to pynbody
        # and filename returned by timestep.filename, removing need for pynbody_path_from_snapdir_path
        # and for this hack.
        #
        # This should probably actually be solved in pynbody.
        if "snap_" in timestep_filename:
            filename = os.path.dirname(timestep_filename)
        else:
            filename = timestep_filename

        stepid = CaterpillarInputHandler._snap_id_from_snapdir_path(filename)
        if stepid:
            listname = "out_%d.list"%stepid
            path = os.path.join(CaterpillarInputHandler._rockstar_path_from_snapdir_path(filename),
                                listname)
            return path
        else:
            return None



import numpy as np
import sys

is_none = np.vectorize(lambda x: x is None, "?")
is_not_none = np.vectorize(lambda x: x is not None, "?")
is_false = np.vectorize(lambda x: x is False, "?")
is_not_false = np.vectorize(lambda x: x is not False, "?")

if sys.version_info[0]<3:
    from ._lru_cache import lru_cache
else:
    from functools import lru_cache



from __future__ import absolute_import
import numpy as np

from .. import BuiltinFunction, FixedInput, FixedNumericInput, StoredProperty, MultiCalculation, ReturnInputHalos
from ... import core
from ... import relation_finding
from ... import temporary_halolist as thl

def _find_progenitor_or_descendant(source_halos, property_proxy, property_criterion, strategy):
    if property_criterion != 'min' and property_criterion != 'max':
        raise ValueError("Property criterion must be either 'min' or 'max'")

    if len(source_halos) == 0:
        return []

    all_major_progs_strategy = strategy(source_halos)

    sources_for_each_value = all_major_progs_strategy.sources()

    property_and_obj = MultiCalculation(ReturnInputHalos(), property_proxy.name)

    with all_major_progs_strategy.temp_table() as tt:
        # the query has to explicitly include the source_id, otherwise the sqlalchemy dedup
        # process removes duplicates rows (e.g. if there are several null results, only the
        # first will be returned!)
        raw_query = thl.enumerated_halo_query(tt)
        query = property_and_obj.supplement_halo_query(raw_query)
        all_major_progs = [x[1] for x in query.all()]
        db_objects_for_each_value, values = property_and_obj.values(all_major_progs)

    assert len(values) == len(all_major_progs) == len(sources_for_each_value)

    # eliminate all None values
    mask = np.asarray(values)!=None
    sources_for_each_value = np.asarray(sources_for_each_value)[mask]
    db_objects_for_each_value = np.asarray(db_objects_for_each_value)[mask]
    values = np.asarray(values)[mask]

    # now re-organize the values so that we have one per source
    values_per_source = {s: [] for s in range(len(source_halos))}
    objs_per_source = {s: [] for s in range(len(source_halos))}
    for source, value, obj in zip(sources_for_each_value, values, db_objects_for_each_value):
        values_per_source[source].append(value)
        objs_per_source[source].append(obj)

    results = []
    for s in range(len(source_halos)):
        vals = values_per_source[s]
        objs = objs_per_source[s]
        assert len(vals) == len(objs)
        if len(vals) == 0:
            results.append(None)
        else:
            try:
                if property_criterion == 'min':
                    index = np.argmin(vals)
                elif property_criterion == 'max':
                    index = np.argmax(vals)
                else:
                    assert False  # should not reach this point
                results.append(objs[index])
            except ValueError:
                results.append(None) # argmin/argmax of empty sequence -> no candidate results

    return results

@BuiltinFunction.register
def find_progenitor(source_halos, property_proxy, property_criterion):
   return _find_progenitor_or_descendant(source_halos, property_proxy, property_criterion,
                                         relation_finding.multi_source.MultiSourceAllMajorProgenitorsStrategy)


@BuiltinFunction.register
def find_descendant(source_halos, property_proxy, property_criterion):
   return _find_progenitor_or_descendant(source_halos, property_proxy, property_criterion,
                                         relation_finding.multi_source.MultiSourceAllMajorDescendantsStrategy)


find_progenitor.set_input_options(0, provide_proxy=True, assert_class = StoredProperty)
find_progenitor.set_input_options(1, provide_proxy=True, assert_class = FixedInput)

find_descendant.set_input_options(0, provide_proxy=True, assert_class = StoredProperty)
find_descendant.set_input_options(1, provide_proxy=True, assert_class = FixedInput)